"""
hook -- Hook based implementation
"""

import collections
import heapq

from searchtree import Node, LinkedTree as Tree


XYZ = collections.namedtuple('XYZ', 'x y z')
Event = collections.namedtuple('Event', 'x y extra_data')
_Segment = collections.namedtuple('Segment', 'x y length')
_SlopeSegment = collections.namedtuple('SlopeSegment', 'x y z length')


class Segment(_Segment):
    """
    Segments live in x-y plane
    """
    def end(self):
        """
        Return segment of length 0 representing end of the segment.
        """
        return Segment(self.x + self.length, self.y + self.length, 0)


class SlopeSegment(_SlopeSegment):
    """
    SlopeSegments live in x-y-z space
    """
    @classmethod
    def init_by_segment(cls, segment, c0):
        """
        Return slope segment based on segment and z-section of y.
        """
        return cls(segment.x,
                   segment.y,
                   segment.y + c0,
                   segment.length)

    def delay(self, diff):
        """
        Return slope segment whose start delays diff from self, but
        end point remains same.
        """
        return SlopeSegment(self.x + diff,
                            self.y + diff,
                            self.z + diff,
                            self.length - diff)

    def shorten(self, diff):
        """
        Return diff shortened slope segment from self; its start point
        remains same.
        """
        return SlopeSegment(self.x,
                            self.y,
                            self.z,
                            self.length - diff)

    def end(self):
        """
        Return an XYZ instance representing end of the segment.
        """
        return XYZ(self.x + self.length,
                   self.y + self.length,
                   self.z + self.length)


class Hook(object):
    """
    Hook stores info of slope in y-z plane based on segments, and
    plateau of upper bound
    """

    # coefficient as polynomial functions
    # (z = y + c0; c0 varies instance by instance)
    def __init__(self, zsection, segment):
        """
        The hook consists of a slope with given z-section and
        a plateau with given upperbound.
        """
        self.c0 = zsection
        self.slopesegments = Tree.get_empty(Node)
        new_segment = SlopeSegment.init_by_segment(segment, self.c0)
        self.slopesegments.insert(segment.x, new_segment)

    def __call__(self, x, y):
        """
        Return z-value at x, y.

        If y is in slope part, the value is determined from the linear
        function.  Otherwise, it is capped with the upper bound
        (plateau).
        """
        corner = self.corner(x)
        if y < corner.y:
            return y + self.c0
        else:
            return corner.z

    def corner(self, x):
        """
        Return 3-dimensional coordinate (x, y, z) of corner of the hook.
        """
        pred, hit, _ = self.slopesegments.search_interval(x)
        if not pred and not hit:
            # pre-start:
            # The y-section of extended (future) slope is returned.
            y_segment = -self.c0
            z_segment = 0
        elif hit:
            y_segment = hit.node.value.y
            z_segment = hit.node.value.z
        else:
            segment = pred.node.value
            end = segment.end()
            if x < end.x:
                y_segment = segment.y + (x - segment.x)
                z_segment = segment.z + (x - segment.x)
            else:
                y_segment = end.y
                z_segment = end.z

        return XYZ(x, y_segment, z_segment)

    def get_max_slopesegment(self):
        """
        Return max of slopesegments.
        """
        return self.slopesegments.max().node.value

    def update_slopesegments(self, segment):
        """
        Update slopesegments by a new segment.

        New segment will be in front of old segments.
        """
        _, hit, succ = self.slopesegments.search_interval(segment.x)
        new_segment = SlopeSegment.init_by_segment(segment, self.c0)
        if not succ:
            # no correlation.
            assert not hit
            self.slopesegments.insert_and_link(segment.x, new_segment)
            return

        # old_segment starts from below segment.x
        cumulative_length = 0
        while succ:
            old_key, old_segment = succ.node.key, succ.node.value
            cumulative_length += old_segment.length
            # remove the current
            self.slopesegments.delete_and_unlink(old_key)
            if new_segment.length < cumulative_length:
                # insert shortened old
                diff = cumulative_length - new_segment.length
                shortened_segment = old_segment.delay(old_segment.length - diff)
                self.slopesegments.insert_and_link(
                    shortened_segment.x, shortened_segment)
                break
            elif new_segment.length == cumulative_length:
                # replaced
                break

            # get next
            _, _, succ = self.slopesegments.search_interval(old_key)

        self.slopesegments.insert_and_link(new_segment.x, new_segment)

    def forget_below(self, x):
        """
        Discard slope segments below x, and return True if something
        remains.
        """
        while self.slopesegments.node.value.end().x <= x:
            # only right subtree remains
            if self.slopesegments.succ:
                self.slopesegments.succ.pred = None
                self.slopesegments = self.slopesegments.right
            else:
                self.slopesegments = Tree.get_empty(Node)
                break

        cell = self.slopesegments.min()
        deleted_keys = []
        on_segment = False
        while cell:
            if cell.node.value.end().x <= x:
                deleted_keys.append(cell.node.key)
                cell = cell.succ
            else:
                if cell.node.value.x < x:
                    on_segment = True
                break

        # delete segments below x
        for key in deleted_keys:
            self.slopesegments.delete_and_unlink(key)

        # delay the segment which x is on
        if on_segment:
            cell = self.slopesegments.min()
            segment = cell.node.value
            self.slopesegments.delete_and_unlink(cell.node.key)
            segment = segment.delay(x - segment.x)
            self.slopesegments.insert_and_link(x, segment)

        return not self.slopesegments.is_bottom()

    def extend(self, segment):
        """
        Extend defined range by given segment, and return the
        extendedly defined new segment.
        """
        old_x = segment.x
        support = self.lookup_support(old_x, None)
        while support is not None:
            next_y = support.end().y
            delay = next_y - segment.y
            if delay >= segment.length:
                # no need to add
                new_segment = Segment(segment.x, segment.y, 0)
                break
            new_x = segment.x + delay
            if old_x == new_x:
                # saturated
                new_segment = Segment(new_x, next_y, segment.length - delay)
                self.update_slopesegments(new_segment)
                break
            old_x = new_x
            support = self.lookup_support(old_x, None)
        else:
            self.update_slopesegments(segment)
            new_segment = segment
        return new_segment

    def lookup_support(self, x, raiseifmiss=True):
        """
        Look up and return a slope segment which includes given x in
        its defining range or just below the x.

        If raiseifmiss is false, return raiseifmiss instead of raising
        LookupError.
        """
        pred, hit, _ = self.slopesegments.search_interval(x)
        if hit:
            return hit.node.value
        elif pred:
            return pred.node.value
        else:
            if not raiseifmiss:
                return raiseifmiss
            raise LookupError("no support found for %d" % x)

    def succ_alive(self, succ, x):
        """
        If succ will be eclipsed by self, then return False.
        Otherwise True.
        """
        diff_z = self.corner(x).z - succ.corner(x).z
        if diff_z >= 0:
            slopesegment = self.lookup_support(x)
            last_x = slopesegment.end().x
            if slopesegment.x == x or last_x > x:
                succ.forget_below(last_x + diff_z)
                if not succ.slopesegments:
                    return False # no succ is needed
        return True

    def recalc_overtake(self, succ, manager):
        """
        Recalculate point to overtake succ, and return as an event.
        """
        pred_max_slopesegment = self.get_max_slopesegment()
        pred_end = pred_max_slopesegment.end()
        succ_max_slopesegment = succ.get_max_slopesegment()
        succ_end = succ_max_slopesegment.end()
        if succ_end.z > pred_end.z:
            return None # nothing to do

        # forget from above if succ is hidden when it stops
        if succ_end.z <= self.corner(succ_end.x).z:
            result_tuple = self._forget_from_above(succ, manager)
            if result_tuple[0]:
                return result_tuple[1]
            else:
                succ_end = result_tuple[1]

        # succ is seen when it reaches its end
        equivalent_x_corner = self.corner(succ_end.x)
        if equivalent_x_corner.z < succ_end.z <= pred_end.z:
            # succ should be overtaken
            if equivalent_x_corner.z == 0:
                support = self.slopesegments.min().node.value
            else:
                support = self.lookup_support(succ_end.x)
            while support.end().z < succ_end.z:
                support = self.slopesegments.search(support.x).succ.node.value
            support_end = support.end()
            diff = support_end.z - succ_end.z
            corner_x, corner_y = support_end.x - diff, support_end.y - diff
            return Event(corner_x, corner_y, (self.c0, succ.c0))

    def _forget_from_above(self, succ, manager):
        """
        Forget slopesegments of succ from above.
        Return (bool, event) where bool is True if the result should be
        the result of caller, too.  If bool is False,
        succ_end.z > self.corner(succ_end.x).z holds.
        """
        succ_max_slopesegment = succ.get_max_slopesegment()
        succ_end = succ_max_slopesegment.end()
        support = self.lookup_support(succ_end.x)
        while succ_end.z <= self.corner(succ_end.x).z:
            assert succ_end.z <= support.end().z
            rest = support.z + support.length - succ_end.z
            diff = succ_end.z - support.z
            if diff >= succ_max_slopesegment.length:
                succ_max_key = succ.slopesegments.max().node.key
                succ.slopesegments.delete_and_unlink(succ_max_key)
                if not succ.slopesegments:
                    event = manager.delete_and_unlink(succ.c0)
                    if event is not None:
                        manager.add_event(event)
                    return (True, None)
                succ_max_slopesegment = succ.get_max_slopesegment()
                succ_end = succ_max_slopesegment.end()
            else:
                if diff > 0:
                    # update succ
                    # note: this should be done before possible return in
                    #       update support in order to keep consistency
                    succ_max_slopesegment = succ_max_slopesegment.shorten(diff)
                    succ_end = succ_max_slopesegment.end()
                    succ.slopesegments.max().node.value = succ_max_slopesegment
                # update support
                possible_event = Event(support.x, support.y, (self.c0, succ.c0))
                support = self._next_support(support.x)
                if support is None:
                    return (True, possible_event)

            # as far as the program consistently treats slopesesegments,
            assert succ_end.z <= support.end().z

        return (False, succ_end)

    def _next_support(self, key):
        """
        Return lower slopesegment than key.
        """
        pre_support, hit, _ = self.slopesegments.search_interval(key)
        assert hit
        return pre_support.node.value
        

class HookNode(Node):
    """
    hook to be used in doubly-linked list tree.
    """
    _bottom_hook =  Hook(0, Segment(0, 0, 0))

    def __init__(self, key, value, level):
        """
        HookNode(zsection, segment/hook, level)

        This rather strange initialization style comes from the
        requirement of Tree class's usage of node.
        """
        if isinstance(value, Hook):
            Node.__init__(self, key, value, level)
        elif isinstance(value, Segment):
            hook = Hook(key, value)
            Node.__init__(self, key, hook, level)
        elif isinstance(value, HookNode):
            assert key == value.key
            assert isinstance(value, Hook), type(value)
            Node.__init__(self, key, value.value, level)
        else:
            raise ValueError("'value' is: %s" % type(value))

    @classmethod
    def get_bottom(cls):
        """
        get bottom.
        """
        return cls(0, cls._bottom_hook, 0)

    def is_bottom(self):
        """
        Return True if the object is identical to bottom
        """
        return self.level == 0

    def __repr__(self):
        return str(self.__dict__)

    # comparisons
    # be careful that the order is reversed for key
    def __lt__(self, other):
        """
        inequality for sort
        """
        if hasattr(other, "key"):
            return self.key > other.key
        else:
            return self.key > other

    def __le__(self, other):
        """
        inequality for sort
        """
        if hasattr(other, "key"):
            return self.key >= other.key
        else:
            return self.key >= other

    def __eq__(self, other):
        """
        inequality for sort
        """
        if hasattr(other, "key"):
            return self.key == other.key
        else:
            return self.key == other

    def __ne__(self, other):
        """
        inequality for sort
        """
        if hasattr(other, "key"):
            return self.key != other.key
        else:
            return self.key != other

    def __ge__(self, other):
        """
        inequality for sort
        """
        if hasattr(other, "key"):
            return self.key <= other.key
        else:
            return self.key <= other

    def __gt__(self, other):
        """
        inequality for sort
        """
        if hasattr(other, "key"):
            return self.key < other.key
        else:
            return self.key < other

    # delegated methods
    def __call__(self, x, y):
        """
        Return z-value at x, y.

        If y is in slope part, the value is determined from the linear
        function.  Otherwise, it is capped with the upper bound
        (plateau).
        """
        return self.value(x, y)

    def __getattr__(self, attr):
        """
        If attr undefined in this class is defined in Hook, then it
        can be accecible as if it is defined in this class.
        """
        if hasattr(self.value, attr):
            return getattr(self.value, attr)
        raise AttributeError("HookNode instance doesn't have '%s' attr" % attr)


class EventManager(object):
    """
    Read information of segments and events, update the function
    envelope with several slopes.
    """
    def __init__(self, string_length):
        """
        EventManager(string_length)
        """
        self.string_length = string_length
        # jar
        self.jar = []
        # envelope
        self.envelope = Tree.get_empty(HookNode)

    def set_data(self, segments):
        """
        Populate the jar.

        segments can be any iterable (including iterator) of triples.
        """
        for triple in segments:
            heapq.heappush(self.jar, Segment(*triple))

    def add_event(self, event):
        """
        Add an event.
        """
        heapq.heappush(self.jar, event)

    def get_next_segments_and_events(self):
        """
        Return segments/events of the same x coordinate with the
        top of the heap.
        If the event jar is empty, return an empty list.
        """
        segments = []
        events = []
        if self.jar:
            top_x = self.jar[0].x
            while self.jar and self.jar[0].x == top_x:
                top = heapq.heappop(self.jar)
                if isinstance(top, Segment):
                    segments.append(top)
                else:
                    events.append(top)
        return (segments, events)

    def update_envelope(self, x):
        """
        Update envelope up to x
        """
        if not self.jar:
            return

        top_x = self.jar[0].x
        while self.jar and top_x <= x:
            segments, events = self.get_next_segments_and_events()

            for event in events:
                self.dispatch_event(event)

            for segment in segments:
                assert segment.length > 0
                self.update_hook_by_segment(segment)

            if self.jar:
                top_x = self.jar[0].x

    def dispatch_event(self, event):
        """
        Call appropriate method of Hook according to the event.
        """
        zsection, overtakee = event.extra_data
        xyz = XYZ(event.x, event.y, event.y + zsection)
        hook = self.envelope.search(zsection)
        if not hook:
            return
        succ = self.envelope.search(overtakee)
        if not succ:
            return
        self.overtake(xyz, succ)

    def overtake(self, xyz, succ):
        """
        Overtake succ.

        Delete succ if needed.
        """
        # deleted slopesegment of succ should be calculated back from
        # xyz
        max_segment = succ.node.slopesegments.max()
        while max_segment is not None and max_segment.node.value.z > xyz.z:
            max_segment = max_segment.pred
        if max_segment is None:
            return
        max_segment = max_segment.node.value
        bound_x = max_segment.x + xyz.z - max_segment.z
        while succ and not succ.node.forget_below(bound_x):
            event = self.delete_and_unlink(succ.node.key)
            if event is not None:
                if event.x <= xyz.x:
                    succ = self.envelope.search(event.extra_data[1])
                else:
                    self.add_event(event)
                    break
            else:
                break

    def update_hook_by_segment(self, segment):
        """
        Update hooks according to the segment.
        """
        zsection = self.get_start_value(segment) - segment.y
        pred, hit, succ = self.envelope.search_interval(zsection)

        if not hit.is_bottom():
            # hook found
            new = self._extend_hook_by_segment(hit, segment)
        else:
            assert linked_or_terminal(pred, succ)
            # new hook
            new = self._create_new_hook_by_segment(pred, segment, zsection)

        if new is None:
            # segment is useless
            return

        # set OVERTAKE if needed
        assert new.node.key == zsection, (new.node.key, zsection)
        pred, _, succ = self.envelope.search_interval(zsection)
        if pred:
            event = pred.node.recalc_overtake(new.node, self)
            if event is not None:
                self.add_event(event)
        if succ:
            event = new.node.recalc_overtake(succ.node, self)
            if event is not None:
                self.add_event(event)

    @staticmethod
    def _extend_hook_by_segment(hit, segment):
        """
        Extend already existing hook (hit) by segment.
        Return the extended hook if segment is useful, None otherwise.
        """
        new_segment = hit.node.value.extend(segment)
        if new_segment.length:
            return hit
        else:
            return None

    def _create_new_hook_by_segment(self, pred, segment, zsection):
        """
        Create a new hook by segment. Its predecessor is pred and it
        has zsection as its key.  Return the created hook if segmnet
        is useful, None otherwise.
        """
        if pred.is_bottom():
            # insert new_hook in tree and link it in doubly-linked list
            self.envelope.insert_and_link(zsection, segment)
            # pick up the new (which is the minimum)
            new = self.envelope.min()
        else:
            new_hook = HookNode(zsection, segment, 1)
            if pred.node.value.succ_alive(new_hook, segment.x):
                # insert new_hook in tree and link it in doubly-linked list
                self.envelope.insert_and_link(zsection, segment)
                # get new by search (insertion breaks the ties that bind
                # variable and NODE, because variables have TREE)
                new = self.envelope.search(zsection)
            else:
                # new_hook is useless
                new = None
        return new

    def delete_and_unlink(self, key):
        """
        Delete node specified by key from the envelope, and recalculate
        overtake point of pred and succ of the deleted node.
        """
        self.envelope.delete_and_unlink(key)
        pred, _, succ = self.envelope.search_interval(key)
        if pred and succ:
            return pred.node.recalc_overtake(succ.node, self)

    def get_start_value(self, segment):
        """
        Get z value of segment at its beginning (x, y)
        """
        self.update_envelope(segment.x) # ensure ready to calculate
        return self._get_start_value_in_subtree(self.envelope, segment)

    def _get_start_value_in_subtree(self, subtree, segment):
        """
        Get z value of segment at its beginning (x, y)
        """
        maximum = 0
        while not subtree.is_bottom():
            hook = subtree.node.value
            maximum = max(maximum, hook(segment.x, segment.y))
            corner = hook.corner(segment.x)
            if corner.y >= segment.y:
                subtree = subtree.left
            else:
                subtree = subtree.right
        return maximum

    def get_result(self):
        """
        Return the final result.

        The final result is envelope's value at self.string_length.
        """
        self.update_envelope(self.string_length)
        probe = Segment(self.string_length, self.string_length, 1)
        return self.string_length - self.get_start_value(probe)


def linked_or_terminal(pred, succ):
    """
    If one of pred or succ is bottom or pred and succ are doubly linked
    to each other, return True.
    """
    if not pred or not succ:
        return True
    return pred.succ == succ and succ.pred == pred
